import { __awaiter, __decorate, __param } from "tslib";
import { Connect, CredentialManager, MessageBuilder } from '@genesislcap/foundation-comms';
import { AbstractMachine, createErrorStateNode, FetchMachine, isDoneInvokeEvent, isErrorEvent, MappingError, setError, } from '@genesislcap/foundation-state-machine';
import { UUID } from '@genesislcap/foundation-utils';
import { DI } from '@microsoft/fast-foundation';
import { and, assign, createMachine, fromPromise, not, sendTo } from 'xstate';
import { User } from '../entity';
import { LOGIN_DETAILS_URL, LOGIN_REFRESH_URL, LOGIN_URL, LOGOUT_URL, SOCKET_URL } from '../env';
import { AuthConfig } from '../main/config';
import { isLoginAckEntity, isLoginNackEntity, isLogoutAckEntity, isLogoutNackEntity, LoginMapper, LogoutMapper, } from '../mapper';
import { logger } from '../utils';
import { isAuthError } from './errors';
import { defaultAuthMachineContext } from './types';
/**
 * Default AuthMachine.
 * @public
 */
let DefaultAuthMachine = class DefaultAuthMachine extends AbstractMachine {
    constructor(user, connect, fetchMachine, messageBuilder, credentialManager, uuid, loginMapper, logoutMapper, config) {
        super();
        this.user = user;
        this.connect = connect;
        this.fetchMachine = fetchMachine;
        this.messageBuilder = messageBuilder;
        this.credentialManager = credentialManager;
        this.uuid = uuid;
        this.loginMapper = loginMapper;
        this.logoutMapper = logoutMapper;
        this.config = config;
        this.machine = createMachine({
            id: 'authMachine',
            context: ({ input }) => (Object.assign({}, input)),
            initial: 'authCheck',
            states: {
                authCheck: {
                    tags: ['authCheck', 'pending'],
                    meta: {
                        content: 'Authentication check, please wait.',
                    },
                    always: [
                        {
                            guard: and(['isAuthenticated', not('isBanned')]),
                            target: 'loggedIn',
                        },
                        {
                            target: 'loggedOut',
                        },
                    ],
                },
                loggingIn: {
                    tags: ['loggingIn', 'pending'],
                    meta: {
                        content: 'Logging in, please wait.',
                    },
                    invoke: {
                        id: 'loginViaFetch',
                        src: fetchMachine.machine,
                        onDone: {
                            actions: ['setAuthDTO'],
                            target: 'mapAuthEntity',
                        },
                        onError: 'error.login',
                        onSnapshot: {
                            actions: ({ event }) => {
                                console.log(event.data);
                            },
                        },
                    },
                    entry: [
                        'setLoginMapper',
                        sendTo('loginViaFetch', ({ context }) => {
                            const { credentials } = context;
                            const message = this.messageBuilder.createLoginMessage(credentials);
                            return {
                                type: 'fetch',
                                params: {
                                    url: LOGIN_URL,
                                    requestInit: {
                                        method: 'POST',
                                        body: JSON.stringify(message),
                                        headers: this.messageBuilder.createHTTPHeadersFromMessage(message),
                                        credentials: 'include',
                                    },
                                },
                            };
                        }),
                    ],
                },
                refreshing: {
                    tags: ['refreshing', 'pending'],
                    meta: {
                        content: 'Refreshing, please wait.',
                    },
                    invoke: {
                        id: 'refreshViaFetch',
                        src: fetchMachine.machine,
                        onDone: {
                            actions: ['setAuthDTO'],
                            target: 'mapAuthEntity',
                        },
                        onError: 'error.refresh',
                    },
                    entry: [
                        'setLoginMapper',
                        sendTo('refreshViaFetch', ({ context }) => {
                            const message = {
                                SOURCE_REF: this.uuid.createId(),
                                // SESSION_AUTH_TOKEN: this.session.getSessionStorageItem('authToken'),
                                // MESSAGE_TYPE: EventMessageType.EVENT_LOGIN_AUTH,
                                // DETAILS: {
                                //   REFRESH_AUTH_TOKEN: this.session.getSessionStorageItem('refreshToken'),
                                // },
                            };
                            return {
                                type: 'fetch',
                                params: {
                                    url: LOGIN_REFRESH_URL,
                                    requestInit: {
                                        method: 'POST',
                                        body: JSON.stringify(message),
                                        headers: this.messageBuilder.createHTTPHeadersFromMessage(message),
                                        credentials: 'include',
                                    },
                                },
                            };
                        }),
                    ],
                },
                loginDetails: {
                    tags: ['loginDetails', 'pending'],
                    meta: {
                        content: 'Getting login details, please wait.',
                    },
                    invoke: {
                        id: 'loginDetailsViaFetch',
                        src: fetchMachine.machine,
                        onDone: {
                            actions: ['setAuthDTO'],
                            target: 'mapAuthEntity',
                        },
                        onError: 'error.loginDetails',
                    },
                    entry: [
                        'setLoginMapper',
                        sendTo('loginDetailsViaFetch', ({ context }) => {
                            const message = {
                                SOURCE_REF: this.uuid.createId(),
                                // SESSION_AUTH_TOKEN: this.session.getSessionStorageItem('authToken'),
                                // MESSAGE_TYPE: EventMessageType.EVENT_LOGIN_AUTH,
                                // DETAILS: {
                                //   REFRESH_AUTH_TOKEN: this.session.getSessionStorageItem('refreshToken'),
                                // },
                            };
                            return {
                                type: 'fetch',
                                params: {
                                    url: LOGIN_DETAILS_URL,
                                    requestInit: {
                                        method: 'POST',
                                        body: JSON.stringify(message),
                                        headers: this.messageBuilder.createHTTPHeadersFromMessage(message),
                                        credentials: 'include',
                                    },
                                },
                            };
                        }),
                    ],
                },
                mapAuthEntity: {
                    tags: ['mapAuthEntity', 'pending'],
                    meta: {
                        content: 'Mapping entity, please wait.',
                    },
                    invoke: {
                        src: 'mapper',
                        input: ({ context }) => ({
                            dto: context.authDTO,
                            mapper: context.mapper.toEntity,
                        }),
                        onDone: {
                            actions: ['setAuthEntity'],
                            target: '#authMachine.entityCheck',
                        },
                        onError: '#authMachine.error.mapping',
                    },
                },
                entityCheck: {
                    tags: ['entityCheck', 'pending'],
                    meta: {
                        content: 'Entity check, please wait.',
                    },
                    always: [
                        {
                            guard: 'isLoginAckEntity',
                            actions: ['setUser'],
                            target: '#authMachine.authCheck',
                        },
                        {
                            guard: 'isLogoutAckEntity',
                            actions: ['unsetUser', 'openOIDCLogoutPage'],
                            target: '#authMachine.authCheck',
                        },
                        {
                            guard: 'isLoginNackEntity',
                            target: 'loginNackCheck',
                        },
                        {
                            guard: 'isLogoutNackEntity',
                            target: '#authMachine.error.logout',
                        },
                        {
                            target: '#authMachine.error.mapping',
                        },
                    ],
                },
                loginNackCheck: {
                    tags: ['loginNackCheck', 'pending'],
                    meta: {
                        content: 'Login nack, check details, please wait.',
                    },
                    always: [
                        {
                            guard: 'isMFARequired',
                            target: '#authMachine.error.mfaRequired',
                        },
                        {
                            target: '#authMachine.error.login',
                        },
                    ],
                },
                connectionCheck: {
                    tags: ['connectionCheck', 'pending'],
                    meta: {
                        content: 'Connection check, please wait.',
                    },
                    always: [
                        {
                            guard: 'isConnected',
                            target: '#authMachine.connected',
                        },
                        {
                            target: '#authMachine.disconnected',
                        },
                    ],
                },
                connected: {
                    tags: ['connected'],
                    meta: {
                        content: 'Connected (NEEDS TO MOVE OUT).',
                    },
                },
                disconnected: {
                    tags: ['disconnected'],
                    meta: {
                        content: 'Disconnected (NEEDS TO MOVE OUT).',
                    },
                },
                loggedIn: {
                    tags: ['loggedIn'],
                    meta: {
                        content: 'Logged in.',
                    },
                    entry: ['storeCredentials', 'loggedInCallback', 'routeToRedirectUrl'],
                },
                loggingOut: {
                    tags: ['loggingOut', 'pending'],
                    meta: {
                        content: 'Logging out, please wait.',
                    },
                    invoke: {
                        id: 'logoutViaFetch',
                        src: fetchMachine.machine,
                        onDone: {
                            actions: ['setAuthDTO'],
                            target: 'mapAuthEntity',
                        },
                        onError: 'error.logout',
                    },
                    entry: [
                        'setLogoutMapper',
                        sendTo('logoutViaFetch', ({ context }) => {
                            const headers = {
                                'Content-type': 'application/json; charset=UTF-8',
                                SOURCE_REF: this.uuid.createId(),
                            };
                            return {
                                type: 'fetch',
                                params: {
                                    url: LOGOUT_URL,
                                    requestInit: {
                                        method: 'POST',
                                        headers,
                                        credentials: 'include',
                                    },
                                },
                            };
                        }),
                    ],
                },
                openSocketTest: {
                    tags: ['openSocketTest', 'pending'],
                    meta: {
                        content: 'openSocketTest, please wait.',
                    },
                    invoke: {
                        id: 'connectActor',
                        src: 'connect',
                        onDone: '#authMachine.connectionCheck',
                        onError: '#authMachine.error.connect',
                    },
                    // invoke: {
                    //   id: 'childActor',
                    //   src: fromObservable(() => interval$),
                    //   onSnapshot: {
                    //     target: 'success',
                    //     guard: ({ event }) => {
                    //       return event.data === 3;
                    //     }
                    //   }
                    // }
                },
                closeSocketTest: {
                    tags: ['closeSocketTest', 'pending'],
                    meta: {
                        content: 'closeSocketTest, please wait.',
                    },
                    invoke: {
                        id: 'connectActor',
                        src: 'disconnect',
                        onDone: '#authMachine.connectionCheck',
                        onError: '#authMachine.error.disconnect',
                    },
                    // invoke: {
                    //   id: 'childActor',
                    //   src: fromObservable(() => interval$),
                    //   onSnapshot: {
                    //     target: 'success',
                    //     guard: ({ event }) => {
                    //       return event.data === 3;
                    //     }
                    //   }
                    // }
                },
                loggedOut: {
                    tags: ['loggedOut'],
                    meta: {
                        content: 'Logged out.',
                    },
                    /**
                     * clear up data points...
                     */
                    entry: ['trackUserPath', 'preventSilentAccess', 'loggedOutCallback', 'routeToRootUrl'],
                },
                error: {
                    initial: 'unknown',
                    entry: ['setError', 'errorCallback'],
                    states: {
                        unknown: createErrorStateNode('unknown', 'Unknown error.'),
                        login: createErrorStateNode('login', 'Login error.'),
                        refresh: createErrorStateNode('refresh', 'Refresh error.'),
                        loginDetails: createErrorStateNode('login-details', 'Getting login details error.'),
                        mfaRequired: createErrorStateNode('mfa-required', 'MFA required error.'),
                        logout: createErrorStateNode('logout', 'Logout error.'),
                        mapping: createErrorStateNode('mapping', 'Mapping error.'),
                        connect: createErrorStateNode('connect', 'Connection error.'),
                        disconnect: createErrorStateNode('disconnect', 'Disconnection error.'),
                    },
                },
            },
            on: {
                login: {
                    actions: ['resetContext', 'setCredentials'],
                    target: '#authMachine.loggingIn',
                },
                refresh: {
                    target: '#authMachine.refreshing',
                },
                ['login-details']: {
                    target: '#authMachine.loginDetails',
                },
                logout: {
                    target: '#authMachine.loggingOut',
                },
                openSocket: {
                    target: '#authMachine.openSocketTest',
                },
                closeSocket: {
                    target: '#authMachine.closeSocketTest',
                },
                '*': {
                    actions: ({ event }) => {
                        logger.debug('root of auth', event);
                    },
                },
            },
        }, {
            actors: {
                login: fromPromise(({ input, system }) => __awaiter(this, void 0, void 0, function* () {
                    logger.debug('system', system);
                    const { credentials } = input;
                    // maybe service uses machine? not sure... or call machine direct
                    // this.service.login(input.credentials, input.loginUrl)
                })),
                logout: fromPromise(({ input }) => __awaiter(this, void 0, void 0, function* () { })),
                mapper: fromPromise(({ input }) => __awaiter(this, void 0, void 0, function* () {
                    const { dto, mapper } = input;
                    try {
                        return mapper(dto);
                    }
                    catch (error) {
                        throw new MappingError(dto); // < what about stack?
                    }
                })),
                // socket: fromObservable(() => {})
                connect: fromPromise(({ input }) => __awaiter(this, void 0, void 0, function* () { return this.connect.connect(SOCKET_URL); })),
                disconnect: fromPromise(({ input }) => __awaiter(this, void 0, void 0, function* () { return this.connect.disconnect(); })),
            },
            actions: {
                setCredentials: assign(({ event }) => {
                    if (event.type === 'login') {
                        return {
                            credentials: event.params.credentials,
                        };
                    }
                }),
                debugSocket: assign(({ event }) => {
                    if (isDoneInvokeEvent(event)) {
                        logger.debug(event);
                    }
                }),
                setAuthDTO: assign(({ event }) => {
                    if (isDoneInvokeEvent(event)) {
                        return {
                            authDTO: event.output.data,
                        };
                    }
                }),
                setAuthEntity: assign(({ event }) => {
                    if (isDoneInvokeEvent(event)) {
                        return {
                            authEntity: event.output,
                        };
                    }
                }),
                setUser: ({ event }) => {
                    if (isDoneInvokeEvent(event)) {
                        this.user.set(event.output);
                    }
                },
                unsetUser: () => this.user.unset(),
                openOIDCLogoutPage: ({ event }) => {
                    if (isDoneInvokeEvent(event) && event.output.oidcLogoutUrl) {
                        window.open(event.output.oidcLogoutUrl, '_self');
                    }
                },
                setError,
                setLoginMapper: assign({
                    mapper: () => this.loginMapper,
                }),
                setLogoutMapper: assign({
                    mapper: () => this.logoutMapper,
                }),
                resetContext: assign(() => this.startingContext),
                preventSilentAccess: () => __awaiter(this, void 0, void 0, function* () {
                    yield this.credentialManager.preventSilentAccess();
                }),
                storeCredentials: ({ context }) => {
                    if (context.credentials.type === 'BASIC') {
                        this.credentialManager.storeCredentials({
                            id: context.credentials.username,
                            name: context.credentials.username,
                            password: context.credentials.password,
                        });
                    }
                },
                routeToRedirectUrl: () => this.config.routeToUrl(this.user.lastPath() || this.config.defaultRedirectUrl),
                routeToRootUrl: () => this.config.routeToUrl('/'),
                trackUserPath: () => this.user.trackPath(true),
                /**
                 * loggedInCallback.
                 * @remarks
                 * Can be overridden by the caller. Alternatively callers can observe machine.context.
                 * @public
                 */
                loggedInCallback: () => {
                    logger.debug('AuthMachine.loggedInCallback invoked. Should be overwritten in most cases.');
                },
                /**
                 * loggedOutCallback.
                 * @remarks
                 * Can be overridden by the caller. Alternatively callers can observe machine.context.
                 * @public
                 */
                loggedOutCallback: () => {
                    logger.debug('AuthMachine.loggedOutCallback invoked. Should be overwritten in most cases.');
                },
                /**
                 * errorCallback.
                 * @remarks
                 * Can be overridden by the caller. Alternatively callers can observe machine.context.
                 * @public
                 */
                errorCallback: () => {
                    logger.debug('AuthMachine.errorCallback invoked. Should be overwritten in most cases.');
                },
            },
            guards: {
                isAuthError: ({ event }) => (isErrorEvent(event) ? isAuthError(event.data) : false),
                isAuthenticated: ({ context }) => { var _a; return (_a = context.user) === null || _a === void 0 ? void 0 : _a.isAuthenticated(); },
                isBanned: ({ context }) => { var _a; return (_a = context.user) === null || _a === void 0 ? void 0 : _a.isBanned(); },
                isLoginAckEntity: ({ event }) => isDoneInvokeEvent(event) ? isLoginAckEntity(event.output) : false,
                isLoginNackEntity: ({ event }) => isDoneInvokeEvent(event) ? isLoginNackEntity(event.output) : false,
                isMFARequired: ({ event }) => isDoneInvokeEvent(event)
                    ? isLoginNackEntity(event.output) && event.output.isMFARequired
                    : false,
                isLogoutAckEntity: ({ event }) => isDoneInvokeEvent(event) ? isLogoutAckEntity(event.output) : false,
                isLogoutNackEntity: ({ event }) => isDoneInvokeEvent(event) ? isLogoutNackEntity(event.output) : false,
                isConnected: () => this.connect.isConnected,
            },
        });
        this.start();
    }
    start(options = {
        input: Object.assign({ user: this.user }, defaultAuthMachineContext),
    }) {
        super.start(options);
    }
    /**
     * AuthMachine is a singleton.
     */
    getFromContainer() {
        return this.container.get(AuthMachine);
    }
};
DefaultAuthMachine = __decorate([
    __param(0, User),
    __param(1, Connect),
    __param(2, FetchMachine),
    __param(3, MessageBuilder),
    __param(4, CredentialManager),
    __param(5, UUID),
    __param(6, LoginMapper),
    __param(7, LogoutMapper),
    __param(8, AuthConfig)
], DefaultAuthMachine);
export { DefaultAuthMachine };
/**
 * AuthMachine DI key.
 * @internal
 */
export const AuthMachine = DI.createInterface((x) => x.singleton(DefaultAuthMachine));
